"""
A test suite for the decorator arguments.

NOTE that this is an atypical use case in that the namespace is defined.
This is not required in the tasks.py module.
"""

from invocate import task, task_namespace


# --- define tasks
@task
def no_decorator_args(c):
    pass


@task(help={'somearg': 'some argument'})
def with_decorator_args(c, somearg):
    pass


@task(namespace=('ns', 'this'))
def with_namespace_as_tuple(c):
    pass


@task(namespace='ns.that')
def with_namespace_as_string(c):
    pass


@task(namespace='ns.the_other', name='updated-name')
def with_namespace_as_string_and_updated_name(c):
    pass


ns = task_namespace()


# --- define tests
def test_no_decorator_args():
    """It should register tasks with no decorator arguments."""
    assert 'no-decorator-args' in ns.task_names


def test_with_decorator_args():
    """It should register tasks with decorator arguments."""
    assert 'with-decorator-args' in ns.task_names


def test_with_namespace_as_tuple():
    """It should register tasks with namespace as tuple."""
    # print(f'ns: {ns.task_names}')
    assert 'ns.this.with-namespace-as-tuple' in ns.task_names


def test_with_namespace_as_string():
    """It should register tasks with namespace as string."""
    assert 'ns.that.with-namespace-as-string' in ns.task_names


def test_with_namespace_as_string_and_updated_name():
    """It should register tasks with namespace as string and updated name."""
    assert 'ns.the-other.updated-name' in ns.task_names



"""
Invoke's own 'binary' entrypoint.

Dogfoods the `program` module.
"""
from types import ModuleType
from typing import Optional, Dict, Any

from invoke import (
    __version__, Program, Collection, Task, CollectionNotFound,
    Exit)
from invoke.config import copy_dict, merge_dicts

from .core import task_namespace


class InvocateCollection(Collection):
    @classmethod
    def from_module(
        cls,
        module: ModuleType,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        loaded_from: Optional[str] = None,
        auto_dash_names: Optional[bool] = None,
    ) -> "Collection":
        """
        Return a new `.Collection` created from ``module``.

        Inspects ``module`` for any `.Task` instances and adds them to a new
        `.Collection`, returning it. If any explicit namespace collections
        exist (named ``ns`` or ``namespace``) a copy of that collection object
        is preferentially loaded instead.

        When the implicit/default collection is generated, it will be named
        after the module's ``__name__`` attribute, or its last dotted section
        if it's a submodule. (I.e. it should usually map to the actual ``.py``
        filename.)

        Explicitly given collections will only be given that module-derived
        name if they don't already have a valid ``.name`` attribute.

        If the module has a docstring (``__doc__``) it is copied onto the
        resulting `.Collection` (and used for display in help, list etc
        output.)

        :param str name:
            A string, which if given will override any automatically derived
            collection name (or name set on the module's root namespace, if it
            has one.)

        :param dict config:
            Used to set config options on the newly created `.Collection`
            before returning it (saving you a call to `.configure`.)

            If the imported module had a root namespace object, ``config`` is
            merged on top of it (i.e. overriding any conflicts.)

        :param str loaded_from:
            Identical to the same-named kwarg from the regular class
            constructor - should be the path where the module was
            found.

        :param bool auto_dash_names:
            Identical to the same-named kwarg from the regular class
            constructor - determines whether emitted names are auto-dashed.

        .. versionadded:: 1.0
        """
        module_name = module.__name__.split(".")[-1]

        def instantiate(obj_name: Optional[str] = None) -> "Collection":
            # Explicitly given name wins over root ns name (if applicable),
            # which wins over actual module name.
            args = [name or obj_name or module_name]
            kwargs = dict(
                loaded_from=loaded_from, auto_dash_names=auto_dash_names
            )
            instance = cls(*args, **kwargs)
            instance.__doc__ = module.__doc__
            return instance

        obj = task_namespace()
        collection = instantiate()
        collection.tasks = collection._transform_lexicon(obj.tasks)
        collection.collections = collection._transform_lexicon(obj.collections)
        collection.default = (
            collection.transform(obj.default) if obj.default else None
        )
        obj_config = copy_dict(obj._configuration)
        if config:
            merge_dicts(obj_config, config)
        collection._configuration = obj_config
        return collection


class InvocateProgram(Program):
    def load_collection(self) -> None:
        """
        Load a task collection based on parsed core args, or die trying.

        .. versionadded:: 1.0
        """
        # NOTE: start, coll_name both fall back to configuration values within
        # Loader (which may, however, get them from our config.)
        start = self.args["search-root"].value
        loader = self.loader_class(  # type: ignore
            config=self.config, start=start
        )
        coll_name = self.args.collection.value
        try:
            module, parent = loader.load(coll_name)
            # This is the earliest we can load project config, so we should -
            # allows project config to affect the task parsing step!
            # TODO: is it worth merging these set- and load- methods? May
            # require more tweaking of how things behave in/after __init__.
            self.config.set_project_location(parent)
            self.config.load_project()
            self.collection = InvocateCollection.from_module(
                module,
                name='',
                loaded_from=parent,
                auto_dash_names=self.config.tasks.auto_dash_names,
            )
        except CollectionNotFound as e:
            raise Exit("Can't find any collection named {!r}!".format(e.name))


program = InvocateProgram(
    name="Invocate",
    binary="invocate",
    binary_names=["invocate"],
    version=__version__,
)


"""Core functionality for Invocate task management."""

import contextlib
import os
from typing import Callable, Optional, Union, List, Tuple, Literal, Dict

import attrs
import invoke
from invoke import Collection

_task_collector = None
_namespace_tree = None
NO_COLLECTION_DEFINED = 'invocate_no_collection_defined'


@attrs.define
class InvocateTask:
    """Represents an Invocate task with its metadata."""
    task: Callable
    name: str


@attrs.define
class TaskNamespace:
    """
    A data structure for the organization of namespaced tasks into
    invoke namespace collections.
    """
    name: Optional[str] = None
    parent: Optional['TaskNamespace'] = None
    children: Union[List['TaskNamespace'], None] = None
    tasks: Optional[List[callable]] = None
    collection: Optional[Collection] = None

    def __attrs_post_init__(self):
        self.children = self.children or []
        self.tasks = self.tasks or []

    @classmethod
    def add_task(
            cls,
            namespace_tuple: Union[
                Tuple[str], Literal['invocate_no_collection_defined']],
            task: Callable) -> None:
        """Add a task to the tree of namespace collections."""
        if namespace_tuple == NO_COLLECTION_DEFINED:
            toplevel = cls._singleton_root()
            if not toplevel.tasks:
                toplevel.tasks = []
            toplevel.tasks.append(task)
            return

        namespace = cls._fetch_or_provision_namespace(namespace_tuple)
        if not namespace.tasks:
            namespace.tasks = []
        namespace.tasks.append(task)
        namespace.collection = None

    @classmethod
    def as_collection(cls) -> Collection:
        """Return the toplevel namespace as an invoke collection with all descendants added."""
        cursor = cls._singleton_root()
        return cursor._as_collection()

    @classmethod
    def _singleton_root(cls) -> 'TaskNamespace':
        global _namespace_tree
        if _namespace_tree:
            return _namespace_tree
        _namespace_tree = cls()
        return _namespace_tree

    @classmethod
    def _fetch_or_provision_namespace(
            cls, namespace_tuple: Tuple[str]) -> 'TaskNamespace':
        cursor = cls._singleton_root()
        for namespace_name in namespace_tuple:
            cursor = cursor._seek_or_create(namespace_name)
        return cursor

    def _as_collection(self):
        if self.collection:
            return self.collection
        self.collection = Collection(self.name) if self.name else Collection()
        if self.children:
            for child in self.children:
                self.collection.add_collection(child._as_collection())
        if self.tasks:
            for task in self.tasks:
                self.collection.add_task(task.task, name=task.name)
        return self.collection

    def _add_child_namespace(self, namespace_name) -> 'TaskNamespace':
        if not self.children:
            self.children = []
        new_child = self.__class__(name=namespace_name, parent=self)
        self.children.append(new_child)
        return new_child

    def _seek_or_create(self, namespace_name: str) -> 'TaskNamespace':
        for child in self.children:
            if child.name == namespace_name:
                return child
        return self._add_child_namespace(namespace_name)


@attrs.define
class InvocateTaskCollector:
    """A data structure for the collection of namespaced tasks."""
    tasks_dict: Optional[Dict[Tuple, List[InvocateTask]]] = None


    def add(self, namespace: Tuple, task: InvocateTask) -> None:
        """Store a task with its menu parent tuple."""
        if not self.tasks_dict:
            self.tasks_dict = {}
        if namespace not in self.tasks_dict:
            self.tasks_dict[namespace] = []
        self.tasks_dict[namespace].append(task)

    @classmethod
    def reset(cls):
        global _task_collector
        print(f'resetting task collector')
        _task_collector = None

    @classmethod
    def singleton(cls) -> 'InvocateTaskCollector':
        """Return the global task collector instance."""
        global _task_collector
        if _task_collector:
            return _task_collector
        _task_collector = cls()
        return _task_collector

    @classmethod
    def toplevel_invoke_namespace(cls) -> Collection:
        """Return the toplevel invoke task namespace collection."""
        instance = cls.singleton()

        if instance.tasks_dict:
            for menu_tuple, tasks in instance.tasks_dict.items():
                for task in tasks:
                    TaskNamespace.add_task(menu_tuple, task)

        return TaskNamespace.as_collection()


class _InvocateTaskDecorator:
    """Internal decorator class for invoke tasks with namespace support."""

    def __init__(self, *args, **kwargs):
        self.namespace = NO_COLLECTION_DEFINED
        if 'namespace' in kwargs:
            if isinstance(kwargs['namespace'], str):
                self.namespace = tuple(kwargs['namespace'].split('.'))
            elif (isinstance(kwargs['namespace'], tuple)
                  or isinstance(kwargs['namespace'], list)):
                self.namespace = kwargs['namespace']
            else:
                raise TypeError(
                    f"Invalid namespace type: {type(kwargs['namespace'])}")
            del kwargs['namespace']

        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        return self._collect(func)

    def _collect(self, func):
        wrapped_func = invoke.tasks.task(func, **self.kwargs)
        name = self.kwargs.get(
            'name') if 'name' in self.kwargs else func.__name__
        task = InvocateTask(task=wrapped_func, name=name)
        InvocateTaskCollector.singleton().add(self.namespace, task)
        return wrapped_func


def task(*args, **kwargs):
    """
    Decorator for invoke tasks with simplified namespace support.

    Usage:
        @task
        def my_task(c):
            pass

        @task(namespace=('env','build'))
        def deploy(c):
            pass

        @task(namespace='env.build'))
        def test_deploy(c):
            pass
    """
    if args:
        func = args[0]
        wrapped_func = invoke.tasks.task(func)
        name = func.__name__
        task = InvocateTask(task=wrapped_func, name=name)
        InvocateTaskCollector.singleton().add(NO_COLLECTION_DEFINED, task)
        return wrapped_func
    else:
        return _InvocateTaskDecorator(**kwargs)


def task_namespace():
    """Return the complete task namespace collection for use with Invoke."""
    return InvocateTaskCollector.toplevel_invoke_namespace()


@contextlib.contextmanager
def change_directory(path):
    """Context manager to temporarily change the current working directory."""
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


"""
Invocate - Enhanced Invoke task management with namespacing support.

This package provides a decorator-based approach to organizing Invoke tasks
with a simplified model of namespacing.
"""

from .core import (
    task,
    task_namespace
)

__version__ = "0.1.0"
__author__ = "Fred McDavid"
__email__ = "fred@frameworklabs.us"

__all__ = [
    "task",
    "task_namespace",
]